@@ -39,41 +39,6 @@ __all__ = [
 T = TypeVar('T')
 
 
-class _FakeNamedTuple(tuple):
-    """
-    This class simulates a Python3.6 NamedTuple
-    instance.
-
-    It has the same hidden fields, so the same
-    loader for the NamedTuple.
-
-    It needs to be created with fields, field_types, field_defaults
-    """
-
-    def __new__(cls, fields):
-        return super(_FakeNamedTuple, cls).__new__(cls, tuple(fields))
-
-    @property
-    def _fields(self):
-        return self[0]
-
-    @property
-    def __annotations__(self):
-        return self[1]
-
-    @property
-    def _field_defaults(self):
-        return self[2]
-
-    def __call__(self, **kwargs):
-        try:
-            return self[3](**kwargs)
-        except TypeError as e:
-            raise TypedloadTypeError(str(e), type_=self[3], value=kwargs)
-        except Exception as e:
-            raise TypedloadException(str(e), type_=self[3], value=kwargs)
-
-
 class Loader:
     """
     A loader object that recursively loads data into
@@ -111,6 +76,15 @@ class Loader:
         handler. When disabled, the exceptions are not raised
         and the condition is considered False.
 
+    uniondebugconflict: Disabled by default
+        When enabled, all the possible types for the unions
+        are evaluated instead of stopping at the first that works.
+        If more than one type in the union works, an error is raised
+        because the types are conflicting and the union might return
+        different types with the same input value.
+        This option makes the loading slower and is only to be used when
+        debugging issues.
+
     mangle_key: Defaults to 'name'
         Specifies which key is used into the metadata dictionaries
         to perform name-mangling.
@@ -193,6 +167,9 @@ class Loader:
         # Which key is used in metadata to perform name mangling
         self.mangle_key = 'name'
 
+        # Fail if multiple types work within the union
+        self.uniondebugconflict = False
+
         # Objects loaded from a string
         self.strconstructed = {
             Path,
@@ -218,10 +195,10 @@ class Loader:
             (is_set, _setload),
             (is_frozenset, _frozensetload),
             (is_namedtuple, _namedtupleload),
-            (is_dataclass, _namedtupleload),
+            (is_dataclass, _dataclassload),
             (is_forwardref, _forwardrefload),
             (is_literal, _literalload),
-            (is_typeddict, _namedtupleload),
+            (is_typeddict, _typeddictload),
             (lambda type_: type_ in {datetime.date, datetime.time, datetime.datetime}, _datetimeload),
             (lambda type_: type_ in self.strconstructed, _strconstructload),
             (is_attrs, _attrload),
@@ -279,19 +256,14 @@ class Loader:
                     type_=type_
                 )
 
-        # Add type to known types, to resolve ForwardRef later on
-        if self.frefs is not None and hasattr(type_, '__name__'):
-            typename = type_.__name__
-            if typename not in self.frefs:
-                self.frefs[typename] = type_
+            # Add type to known types, to resolve ForwardRef later on
+            if self.frefs is not None and hasattr(type_, '__name__'):
+                typename = type_.__name__
+                if typename not in self.frefs:
+                    self.frefs[typename] = type_
 
         func = self.handlers[index][1]
 
-        if self.dictequivalence:
-            # Convert argparse.Namespace to dictionary
-            if hasattr(value, '_get_kwargs'):
-                value = {k: v for k,v in value._get_kwargs()}
-
         try:
             return func(self, value, type_)
         except Exception as e:
@@ -381,6 +353,9 @@ def _dictload(l: Loader, value, type_) -> Dict:
     Recursively loads both keys and values.
     """
     key_type, value_type = type_.__args__
+
+    value = _dictequivalence(l, value)
+
     try:
         return {
             l.load(k, key_type, annotation=Annotation(AnnotationType.KEY, k)): l.load(v, value_type, annotation=Annotation(AnnotationType.VALUE, v))
@@ -457,53 +432,65 @@ def _mangle_names(namesmap: Dict[str, str], value: Dict[str, Any], failonextra:
     return r
 
 
-def _namedtupleload(l: Loader, value: Dict[str, Any], type_) -> Any:
+def _dataclassload(l: Loader, value: Dict[str, Any], type_) -> Any:
     """
     This loads a Dict[str, Any] into a NamedTuple.
     """
-    if not hasattr(type_, '__dataclass_fields__'):
-        fields = set(type_.__annotations__.keys())
-        optional_fields = set(getattr(type_, '_field_defaults', {}).keys())
-        type_hints = type_.__annotations__
-    else:
-        #dataclass
-        import dataclasses
-        fields = set(type_.__dataclass_fields__.keys())
-        optional_fields = {k for k,v in type_.__dataclass_fields__.items() if
-                           v.init == False or
-                           not isinstance(v.default, dataclasses._MISSING_TYPE) or
-                           not isinstance(v.default_factory, dataclasses._MISSING_TYPE)}
-        type_hints = {k: v.type for k,v in type_.__dataclass_fields__.items()}
+    from dataclasses import _MISSING_TYPE as DT_MISSING_TYPE
+    fields = set(type_.__dataclass_fields__.keys())
+    necessary_fields = {k for k,v in type_.__dataclass_fields__.items() if
+                        v.init == True and
+                        isinstance(v.default, DT_MISSING_TYPE) and
+                        isinstance(v.default_factory, DT_MISSING_TYPE)}
+    type_hints = {k: v.type for k,v in type_.__dataclass_fields__.items()}
 
-        #Name mangling
+    #Name mangling
 
-        # Prepare the list of the needed name changes
-        transforms = {}  # type: Dict[str, str]
-        for pyname in fields:
-            if type_.__dataclass_fields__[pyname].metadata:
-                name = type_.__dataclass_fields__[pyname].metadata.get(l.mangle_key)
-                if name:
-                    transforms[name] = pyname
+    # Prepare the list of the needed name changes
+    transforms = {}  # type: Dict[str, str]
+    for pyname in fields:
+        if type_.__dataclass_fields__[pyname].metadata:
+            name = type_.__dataclass_fields__[pyname].metadata.get(l.mangle_key)
+            if name:
+                transforms[name] = pyname
 
-        try:
-            value = _mangle_names(transforms, value, l.failonextra)
-        except ValueError as e:
-            raise TypedloadValueError(str(e), value=value, type_=type_)
+    try:
+        value = _mangle_names(transforms, value, l.failonextra)
+    except ValueError as e:
+        raise TypedloadValueError(str(e), value=value, type_=type_)
+    except AttributeError as e:
+        raise TypedloadAttributeError(str(e), value=value, type_=type_)
 
-    if hasattr(type_, '__required_keys__') and hasattr(type_, '__optional_keys__'):
-        # TypedDict, since 3.9
-        necessary_fields = type_.__required_keys__
-        optional_fields = type_.__optional_keys__
-    elif getattr(type_, '__total__', True) == False:
-        # TypedDict, only for 3.8
-        necessary_fields = set()
-    else:
-        necessary_fields = fields.difference(optional_fields)
+    return _objloader(l, fields, necessary_fields, type_hints, value, type_)
 
+
+def _dictequivalence(l: Loader, value: Any) -> Dict:
+    '''
+    Helper function to convert classes that are functionally
+    the same as a dict into a dict.
+
+    At the moment the only class that can be converted is
+    argparse.Namespace.
+
+    in all other cases this simply returns value.
+    '''
+    # Convert argparse.Namespace to dictionary
+    if l.dictequivalence and hasattr(value, '_get_kwargs'):
+        return {k: v for k,v in value._get_kwargs()}
+    return value
+
+
+def _objloader(l: Loader, fields: Set[str], necessary_fields: Set[str], type_hints, value: Dict[str, Any], type_) -> Any:
     try:
         vfields = set(value.keys())
     except AttributeError as e:
-        raise TypedloadAttributeError(str(e), value=value, type_=type_)
+        newvalue = _dictequivalence(l, value)
+
+        if newvalue is value:
+            raise TypedloadAttributeError(str(e), value=value, type_=type_)
+        else:
+            value = newvalue
+            vfields = set(value.keys())
 
     if necessary_fields.intersection(vfields) != necessary_fields:
         raise TypedloadValueError(
@@ -541,6 +528,37 @@ def _namedtupleload(l: Loader, value: Dict[str, Any], type_) -> Any:
         raise TypedloadValueError(e)
 
 
+def _namedtupleload(l: Loader, value: Dict[str, Any], type_) -> Any:
+    """
+    This loads a Dict[str, Any] into a NamedTuple.
+    """
+    fields = set(type_.__annotations__.keys())
+    optional_fields = set(getattr(type_, '_field_defaults', {}).keys())
+    type_hints = type_.__annotations__
+    necessary_fields = fields.difference(optional_fields)
+
+    return _objloader(l, fields, necessary_fields, type_hints, value, type_)
+
+
+def _typeddictload(l: Loader, value: Dict[str, Any], type_) -> Any:
+    """
+    This loads a Dict[str, Any] into a NamedTuple.
+    """
+    fields = set(type_.__annotations__.keys())
+    type_hints = type_.__annotations__
+
+    if hasattr(type_, '__required_keys__') and hasattr(type_, '__optional_keys__'):
+        # TypedDict, since 3.9
+        necessary_fields = type_.__required_keys__
+    elif getattr(type_, '__total__', True) == False:
+        # TypedDict, only for 3.8
+        necessary_fields = set()
+    else:
+        necessary_fields = fields
+
+    return _objloader(l, fields, necessary_fields, type_hints, value, type_)
+
+
 def _unionload(l: Loader, value, type_) -> Any:
     """
     Loads a value into a union.
@@ -573,17 +591,37 @@ def _unionload(l: Loader, value, type_) -> Any:
             sorted_args.append(t)
 
     # Try all types
+    loaded_count = 0
+    r = None
     for t in sorted_args:
         try:
-            return l.load(value, t, annotation=Annotation(AnnotationType.UNION, t))
+            r = l.load(value, t, annotation=Annotation(AnnotationType.UNION, t))
+            loaded_count += 1
+            if not l.uniondebugconflict:
+                # Do not try more if we are not debugging
+                break
         except Exception as e:
             exceptions.append(e)
-    raise TypedloadValueError(
-        'Value of %s could not be loaded into %s' % (tname(value_type), tname(type_)),
-            value=value,
-            type_=type_,
-            exceptions=exceptions
-    )
+
+    if loaded_count == 1:
+        # Loaded only once, all good
+        return r
+    elif loaded_count == 0:
+        # Could not be loaded
+        raise TypedloadValueError(
+            'Value of %s could not be loaded into %s' % (tname(value_type), tname(type_)),
+                value=value,
+                type_=type_,
+                exceptions=exceptions
+        )
+    else:
+        # Loaded more than once, conflict
+        raise TypedloadTypeError(
+            'Value of %s could be loaded into %s %d times' % (tname(value_type), tname(type_), loaded_count),
+                value=value,
+                type_=type_,
+                exceptions=exceptions
+        )
 
 
 def _enumload(l: Loader, value, type_) -> Enum:
@@ -642,18 +680,16 @@ def _datetimeload(l: Loader, value, type_) -> Union[datetime.date, datetime.time
 
 
 def _attrload(l, value, type_):
-    if not isinstance(value, dict):
-        raise TypedloadTypeError('Expected dictionary, got %s' % tname(type(value)), type_=type_, value=value)
-    value = value.copy()
-    names = []
-    defaults = {}
-    types = {}
+    from attr._make import _Nothing as NOTHING
+
+    fields = {i.name for i in type_.__attrs_attrs__}
+    necessary_fields = set()
+    type_hints = {i.name: i.type for i in type_.__attrs_attrs__}
     namesmap = {}  # type: Dict[str, str]
 
     for attribute in type_.__attrs_attrs__:
-        names.append(attribute.name)
-        types[attribute.name] = attribute.type
-        defaults[attribute.name] = attribute.default
+        if attribute.default is NOTHING and attribute.init:
+            necessary_fields.add(attribute.name)
 
         # Manage name mangling
         if l.mangle_key in attribute.metadata:
@@ -663,20 +699,15 @@ def _attrload(l, value, type_):
         value = _mangle_names(namesmap, value, l.failonextra)
     except ValueError as e:
         raise TypedloadValueError(str(e), value=value, type_=type_)
+    except AttributeError as e:
+        raise TypedloadAttributeError(str(e), value=value, type_=type_)
 
-    t = _FakeNamedTuple((
-        tuple(names),
-        types,
-        defaults,
-        type_,
-    ))
-
-    return _namedtupleload(l, value, t)
+    return _objloader(l, fields, necessary_fields, type_hints, value, type_)
 
 
 def _strconstructload(l: Loader, value, type_):
     """
-    Loader for all the types in the stdlib ipaddress module.
+    Loader for all the types taking a string as single constructor parameter
     """
     try:
         return type_(value)
